// ################################################################################  
// ## Redistributions of source code must retain the above copyright notice
// ## DATE: 2011-03-03  
// ## AUTHOR: SecPoint ApS  
// ## MAIL: info@secpoint.com  
// ## SITE: http://www.secpoint.com  
// ## Other scanners: http://www.secpoint.com/free-port-scan.html
// ## Other scanners: http://www.secpoint.com/free-vulnerability-scan.html
// ##
// ## LICENSE: BSD http://www.opensource.org/licenses/bsd-license.php
// ################################################################################  
// ## 1.0 initial release 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#include <sys/times.h>
#endif
#include "net.h"

#define SLEEP_TIME_MSEC			5
#define ARP_TABLE_FILENAME		"./arp_table.txt"
#define OUI_FILENAME			"./oui.txt"
#define PORT_RES_FILENAME		"./port-numbers.txt"

#define OPTIONS_ALL_PORTS		0x00000001
#define OPTIONS_RESOLVE_PORTS	0x00000002
#define OPTIONS_RESOLVE_MAC		0x00000004
#define OPTIONS_ONFLY_HIDE		0x00000008

typedef struct _port_instance
{
	struct _port_instance *next;
	struct _port_instance *prev;
	unsigned short port;
} port_instance;

static int options = 0;
static unsigned long wait_time = 1000; // msec
static unsigned long num_instances = 10;
static unsigned long IP;
static unsigned short start_port = 1;
static unsigned short end_port = 1024;

static unsigned short current_port;
static unsigned short total_opened;
static unsigned short total_scanned;
static unsigned short total_ports;
static char *port_res_file;
static port_instance *oports = NULL;

static unsigned long get_time()
{
#ifdef WIN32
	return GetTickCount();
#else
	struct tms tm;
	return (unsigned long)times(&tm)*10;
#endif
}

static void argument_error()
{
	printf("Argument error\n");
	exit(0);
}

static void parse_arguments(int argc, char *argv[])
{
	int i;
	char *p, *t;
	
	IP = net_get_iip(argv[1]);
	if (IP == 0 || IP == 0xFFFFFFFF)
	{
		printf("Incorrect IP address: %s\n", argv[1]);
		exit(0);
	}
	for (i = 2; i < argc; i++)
	{
		if (argv[i][0] != '-') continue;
		switch (argv[i][1])
		{
		case 'p':
			if (argv[i][2] != 0) p = &argv[i][2];
			else
			{
				i++;
				if (i == argc) argument_error();
				p = argv[i];
			}
			t = p;
			while (*t != '-') 
			{
				t++;
				if (*t == 0) argument_error();
			}
			*t = 0; t++;
			start_port = atoi(p);
			end_port = atoi(t);
			if (start_port == 0 || end_port == 0 || end_port < start_port)
			{
				printf("Incorrect ports, starting = %d, ending = %d\n", start_port, end_port);
				exit(0);
			}
			break;
		case 'n':
			if (argv[i][2] != 0) p = &argv[i][2];
			else
			{
				i++;
				if (i == argc) argument_error();
				p = argv[i];
			}
			num_instances = atoi(p);
			break;
		case 'w':
			if (argv[i][2] != 0) p = &argv[i][2];
			else
			{
				i++;
				if (i == argc) argument_error();
				p = argv[i];
			}
			wait_time = atoi(p);
			break;
		case 'a':
			options |= OPTIONS_ALL_PORTS;
			break;
		case 'r':
			options |= OPTIONS_RESOLVE_PORTS;
			break;
		case 'M':
			options |= OPTIONS_RESOLVE_MAC;
			break;
		case 'h':
			options |= OPTIONS_ONFLY_HIDE;
			break;
		default:
			break;
		}
	}
}

static void print_general()
{
puts("SecPoint Port scanner tool   (c) www.SecPoint.com\nv. 1.0\n"); // todo: add your text here
}

static void print_help()
{	
	puts("Usage: IP [options]\n");
	puts("Options:");
	puts("\t-p <start>-<end>\tStarting and ending port\n\t\t\t(default: 1-1024)");
	puts("\t-n <inst>\tNumber of simultaneous port scan instances\n\t\t\t(default: 10)"); 
	puts("\t-w <time>\tMax waiting time for each port in milliseconds\n\t\t\t(default: 1000)");
	puts("\t-a\t\tReport all ports (OPENED, CLOSED and TIMEDOUT)");
	puts("\t-r\t\tResolve ports to known services");
	puts("\t-M\t\tMAC lookup of remote IP");
	puts("\t-h\t\tHide on-fly port displaying (display at the end)\n");
	printf("Command line example: 192.168.1.1 -p 1-65535 -n 20 -w 2000 -r -M\n\n");
}

static void print_config()
{
	char *yes = "YES", *no = "NO";
	printf("Scanning IP: %s\n", net_get_aip(IP));
	printf("Port range: %u - %u\n", start_port, end_port);
	printf("Instances: %u, waiting time: %u\n", (unsigned int)num_instances, (unsigned int)wait_time);
	printf("Report all ports: %s, Resolve ports: %s\nMAC lookup: %s, Hide on-fly port displaying: %s\n\n", 
		options & OPTIONS_ALL_PORTS ? yes : no, 
		options & OPTIONS_RESOLVE_PORTS ? yes : no,
		options & OPTIONS_RESOLVE_MAC ? yes : no,
		options & OPTIONS_ONFLY_HIDE ? yes : no);
}

static void print_status()
{
//	printf("Scanned: %u/%u (%d%%), opened: %u\r", total_scanned, total_ports, (total_scanned * 100 / total_ports), total_opened);
//	printf(" opened: %u\r", total_opened);
#ifndef WIN32
	fflush(stdout);
#endif
}

static void print_open_port(unsigned short port, char *service_str)
{
//	if (!service_str) printf("%u tcp open                                 \n", port);
	if (!service_str) printf("%u tcp open\n", port);
	else
	{
		int i, len = 60 - strlen(service_str);
		printf("%u/tcp open  %s", port, service_str);
		for (i = 0; i < len; i++) 
		printf(" ");
		printf("\n");
	}
}

static void process_open_port(unsigned short port)
{
	if (options & OPTIONS_RESOLVE_PORTS)
	{
		char buff[16], *p;
		sprintf(buff, "%u/tcp", port);
		p = strstr(port_res_file, buff);
		if (!p) print_open_port(port, NULL);
		else
		{
			char *t, tmp;
			int len = strlen(buff);
			p += len;
			t = p;
			while (*t == ' ') t++;
			p = t;
			while (*t != '\n') t++;
			tmp = *t;
			*t = 0;
			print_open_port(port, p);
			*t = tmp;
		}
	}
	else print_open_port(port, NULL);
}

static void save_open_port(unsigned short port)
{
	port_instance *pi, *ps = oports;

	pi = (port_instance *)malloc(sizeof(port_instance));
	pi->port = port;

	if (ps)
	{
		while (ps->next)
		{
			if (ps->port > port) break;
			ps = ps->next;
		}

		if (!ps->next && ps->port < port)
		{
			ps->next = pi;
			pi->prev = ps;
			pi->next = NULL;
			return;
		}
		
		if (!ps->prev)
		{
			ps->prev = pi;
			pi->next = ps;
			pi->prev = NULL;
			oports = pi;
		}
		else
		{
			pi->prev = ps->prev;
			pi->prev->next = pi;
			pi->next = ps;
			ps->prev = pi;
		}
	}
	else 
	{
		oports = pi;
		oports->next = NULL;
		oports->prev = NULL;
	}
}

static void on_response(port_response resp, unsigned short port)
{
	total_scanned++;

	switch (resp)
	{
	case CLOSED:
		if (options & OPTIONS_ALL_PORTS && !(options & OPTIONS_ONFLY_HIDE)) 
			printf("%u/tcp closed\n", port);
		break;
	case OPENED:
		if (options & OPTIONS_ONFLY_HIDE) save_open_port(port);
		else process_open_port(port);
		total_opened++;
		break;
	case TIMEDOUT:
		if (options & OPTIONS_ALL_PORTS && !(options & OPTIONS_ONFLY_HIDE)) 
			printf("%u/tcp timedout\n", port);
		break;
	default:
		break;
	}

	if (current_port != 0 && current_port <= end_port)
	{
		net_add(IP, current_port, get_time());
		current_port++;
	}

	print_status();

	return;
}

void scanner()
{
	unsigned long i, time;

	net_init(on_response, wait_time);

	time = get_time();
	total_opened = 0;
	total_scanned = 0;
	total_ports = end_port - start_port + 1;
	current_port = start_port;
	for (i = 0; i < num_instances; i++)
	{
		net_add(IP, current_port, get_time());
		current_port++;
		if (current_port > end_port || current_port == 0) break;
	}

	print_status();

	while (net_check(get_time())) 
	{
#ifdef WIN32
		Sleep(SLEEP_TIME_MSEC);
#else
		usleep(SLEEP_TIME_MSEC * 1000);
#endif
	}
	
	time = get_time() - time;
	printf("\n\nScan time: %u.%03u seconds\n", (unsigned int)(time / 1000), (unsigned int)(time % 1000));

	net_uninit();
	if (options & OPTIONS_ONFLY_HIDE)
	{
		port_instance *tmp, *pi = oports;
		while (pi)
		{
			tmp = pi->next;
			process_open_port(pi->port);
			free(pi);
			pi = tmp;
		}
	}
	printf("\n\n");
}

char *read_file(char *file)
{
	FILE *f;
	long size;
	char *buff;
	int r;

	f = fopen(file, "rb");
	if (!f)
	{
		printf("Failed to open %s file\n", file);
		exit(0);
	}

	fseek(f, 0, SEEK_END);
	size = ftell(f);
	fseek(f, 0, SEEK_SET);
	buff = (char *)malloc((unsigned int)size);

	r = fread(buff, 1, (unsigned int)size, f);
	fclose(f);

	return buff;
}

void resolve_MAC(char *mac)
{
	int i;
	char *p, *b = read_file(OUI_FILENAME);
	mac[8] = 0;
	for (i = 0; i < 8; i++) if (mac[i] >= 'a' && mac[i] <= 'f') mac[i] &= 0xDF;
	p = strstr(b, mac);
	if (p)
	{
		char *t;
		p += 18;
		t = p;
		while (*t != '\n') t++;
		*t = 0;
		printf(" %s\n", p);
	}
	else printf("Adapter manufacturer: UNKNOWN\n");
	free(b);
}

void print_MAC()
{
	int r;
	char buff[256], ips[16], *p, *b;

	strcpy(ips, net_get_aip(IP));

#ifdef WIN32
	sprintf(buff, "arp -a %s > "ARP_TABLE_FILENAME, ips);
#else
	sprintf(buff, "arp -a | grep %s | awk '{ print $4 }' > "ARP_TABLE_FILENAME, ips);
#endif
	r = system(buff);

	b = read_file(ARP_TABLE_FILENAME);
#ifdef WIN32
	p = strstr(b, ips);
#else
	for (p = b; *p; p++) 
	{
		if (*p == ':') *p = '-';
		if (*p == '\n') *p = 0;
	}
	p = b;
#endif
	if (!p) printf("ARP table contains no information about IP: %s\nMAC address cannot be resolved!\n", ips);
	else
		{
#ifdef WIN32
		char *t;
		while (*p != ' ') p++;
		t = p;
		while (*t == ' ') t++;
		p = t;
		while (*t != ' ') t++;
		*t = 0;
#endif
		printf("MAC Address: %s", p);
		resolve_MAC(p);
	}
	free(b);

#ifdef WIN32
	system("erase "ARP_TABLE_FILENAME);
#else
	r = system("rm "ARP_TABLE_FILENAME);
#endif
}

int main(int argc, char *argv[])
{
	print_general();

	if (argc < 2)
	{
		print_help();
		return 0;
	}

	parse_arguments(argc, argv);

	print_config();

	if (options & OPTIONS_RESOLVE_PORTS) port_res_file = read_file(PORT_RES_FILENAME);
	scanner();
	if (options & OPTIONS_RESOLVE_PORTS) free(port_res_file);
	if (options & OPTIONS_RESOLVE_MAC) print_MAC();

	return 0;
}



